@@ -1,7 +1,7 @@
 /*
-* libslack - http://libslack.org/
+* libslack - https://libslack.org
 *
-* Copyright (C) 1999-2001 raf <raf@raf.org>
+* Copyright (C) 1999-2002, 2004, 2010, 2020-2023 raf <raf@raf.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
@@ -14,47 +14,25 @@
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
-* along with this program; if not, write to the Free Software
-* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-* or visit http://www.gnu.org/copyleft/gpl.html
+* along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20010215 raf <raf@raf.org>
+* 20230819 raf <raf@raf.org>
 */
 
 #ifndef LIBSLACK_HDR_H
 #define LIBSLACK_HDR_H
 
-#undef _start_decls
+#undef _begin_decls
 #undef _end_decls
-#undef _args
-#undef _throw
-#undef const
 
 #ifdef __cplusplus
-#define _start_decls extern "C" {
+#define _begin_decls extern "C" {
 #define _end_decls }
 #else
-#define _start_decls
+#define _begin_decls
 #define _end_decls
 #endif
 
-#define _throw
-#if 0
-#ifdef __GNUC__
-#if __cplusplus && ((__GNUC__) >= 3 || (__GNUC_MINOR__) >= 8)
-#undef _throw
-#define _throw throw ()
-#endif
-#endif
-#endif
-
-#if defined __STDC__ || defined __cplusplus
-#define _args(args) args _throw
-#else
-#define _args(args) ()
-#define const
-#endif
-
 #endif
 
 /* vi:set ts=4 sw=4: */
