@@ -1,7 +1,7 @@
 /*
-* libslack - http://libslack.org/
+* libslack - https://libslack.org
 *
-* Copyright (C) 1999-2002 raf <raf@raf.org>
+* Copyright (C) 1999-2002, 2004, 2010, 2020-2023 raf <raf@raf.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
@@ -14,11 +14,9 @@
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
-* along with this program; if not, write to the Free Software
-* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-* or visit http://www.gnu.org/copyleft/gpl.html
+* along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20020916 raf <raf@raf.org>
+* 20230819 raf <raf@raf.org>
 */
 
 /*
@@ -54,33 +52,54 @@ I<libslack(prop)> - program properties file module
 This module provides support for system-wide and user-specific (generic and
 program-specific) properties in "well-known" locations:
 
-    /etc/properties/app             - system-wide, generic properties
-    $HOME/.properties/app           - user-defined, generic properties
-    /etc/properties/app.progname    - system-wide, program-specific properties
-    $HOME/.properties/app.progname  - user-defined, program-specific properties
+    /etc/properties/app          - system-wide, generic properties
+    ~/.properties/app            - user-defined, generic properties
+    /etc/properties/app.progname - system-wide, program-specific properties
+    ~/.properties/app.progname   - user-defined, program-specific properties
 
-When the client first requests, sets or unsets a property, all properties
+Note that, depending on how I<libslack> was installed, the system-wide
+C</etc/properties> directory might not be under C</etc>. It might be under
+another directory, such as C</usr/local/etc> or C</opt/local/etc>. If you
+aren't sure where it is expected, run C<libslack-config --cflags> and look
+for the definition of C<ETC_DIR>.
+
+When the client first requests, sets, or unsets a property, all properties
 relevant to the current program are loaded from these files in the order
 given above. This order ensures that program-specific properties override
-generic properties and user-defined properties override system-wide
+generic properties and that user-defined properties override system-wide
 properties. The client can change properties at runtime and save the current
 properties back to disk (to the user-defined, program-specific properties
 file).
 
 Program names (as returned by I<prog_name(3)>) are converted into file name
 suffixes by replacing every occurrence of the file path separator (C<'/'>)
-with a C<'-'>. Properties files consist of one property per line. Each
-property is specified by its name, followed by C<'='> followed by its value.
-The name must not have a C<'='> in it unless it is quoted with a C<'\'>. The
+with a C<'-'>. Properties files consist of one property per line.
+
+Each property is specified by its name, followed by C<'='> followed by its
+value. The name must not have a C<'='> in it unless it is quoted with a
+preceding C<'\'>. Special characters are expressed in I<C> string literal
+notation (e.g. C<\a\b\f\n\r\t\v\x1b>). Spaces immediately before and after
+the C<'='> are stripped unless they are quoted with a preceding C<'\'>. The
 properties files may also contain blank lines and comments (C<'#'> until the
 end of the line).
 
+Boolean property values can be expressed as C<0> or C<1>, C<true> or
+C<false>, C<yes> or C<no>, C<on> or C<off> (case insensitive).
+
 =over 4
 
 =cut
 
 */
 
+#ifndef _BSD_SOURCE
+#define _BSD_SOURCE /* For snprintf() on OpenBSD-4.7 */
+#endif
+
+#ifndef _DEFAULT_SOURCE
+#define _DEFAULT_SOURCE /* New name for _BSD_SOURCE */
+#endif
+
 #include "config.h"
 #include "std.h"
 
@@ -105,6 +124,10 @@ end of the line).
 #define isodigit(c) (isdigit((int)(unsigned char)c) && (unsigned char)(c) < '8')
 #endif
 
+#ifndef is_space
+#define is_space(c) isspace((int)(unsigned char)(c))
+#endif
+
 typedef struct Prop Prop;
 
 struct Prop
@@ -170,7 +193,7 @@ static void prop_release(Prop *prop)
 
 C<int key_cmp(const char **a, const char **b)>
 
-Compares the two strings pointed to by C<a> and C<b>.
+Compares the two strings pointed to by C<a> and C<b> using I<strcmp(3)>.
 
 */
 
@@ -190,8 +213,10 @@ C<String *quote_special(const char *src)>
 Replaces every occurrence in C<src> of special characters (represented in
 I<C> by C<"\a">, C<"\b">, C<"\f">, C<"\n">, C<"\r">, C<"\t">, C<"\v">) with
 their corresponding I<C> representation. Other non-printable characters are
-replaced with their ASCII codes in hexadecimal (i.e. "\xhh"). The caller
-must deallocate the memory returned.
+replaced with their I<ASCII> codes in hexadecimal (i.e. "\xhh"). It is the
+caller's responsibility to deallocate the returned I<String> with
+I<str_release(3)> or I<str_destroy(3)>. It is strongly recommended to use
+I<str_destroy(3)>, because it also sets the pointer variable to C<null>.
 
 */
 
@@ -207,8 +232,10 @@ C<String *unquote_special(const char *src)>
 Replaces every occurrence in C<src> of C<"\a">, C<"\b">, C<"\f">, C<"\n">,
 C<"\r">, C<"\t"> or C<"\v"> with the corresponding special characters (as
 interpreted by I<C>). Ascii codes in octal or hexadecimal (i.e. "\ooo" or
-"\xhh") are replaced with their corresponding ASCII characters. The caller
-must deallocate the memory returned.
+"\xhh") are replaced with their corresponding I<ASCII> characters. It is the
+caller's responsibility to deallocate the returned I<String> with
+I<str_release(3)> or I<str_destroy(3)>. It is strongly recommended to use
+I<str_destroy(3)>, because it also sets the pointer variable to C<null>.
 
 */
 
@@ -221,8 +248,10 @@ static String *unquote_special(const char *src)
 
 C<String *quote_equals(const char *src)>
 
-Replace every occurrence in C<src> of C<"="> with C<"\=">. The caller must
-deallocate the memory returned.
+Replaces every occurrence in C<src> of C<"="> with C<"\=">. It is the
+caller's responsibility to deallocate the returned I<String> with
+I<str_release(3)> or I<str_destroy(3)>. It is strongly recommended to use
+I<str_destroy(3)>, because it also sets the pointer variable to C<null>.
 
 */
 
@@ -235,8 +264,10 @@ static String *quote_equals(const char *src)
 
 C<String *quote_equals(const char *src)>
 
-Replace every occurrence in C<src> of C<"\="> with C<"=">. The caller must
-deallocate the memory returned.
+Replaces every occurrence in C<src> of C<"\="> with C<"=">. It is the
+caller's responsibility to deallocate the returned I<String> with
+I<str_release(3)> or I<str_destroy(3)>. It is strongly recommended to use
+I<str_destroy(3)>, because it also sets the pointer variable to C<null>.
 
 */
 
@@ -250,7 +281,7 @@ static String *unquote_equals(const char *src)
 C<char *user_home(void)>
 
 Returns the user's home directory (obtained from C</etc/passwd>). The return
-value is cached so any subsequent calls are faster.
+value is cached so that any subsequent calls will be faster.
 
 */
 
@@ -272,12 +303,12 @@ static char *user_home(void)
 
 C<void prop_parse(Map *map, const char *path, char *line, size_t lineno)>
 
-Parses a line from a properties file. C<path> if the path of the properties
-file. C<line> is the text to parse. C<lineno> is the current line number.
-The property parsed, if any, is added to C<map>. Emits error messages when
-syntax errors occur. That's probably a mistake. To suppress the error
-messages, call I<prog_err_none(3)> first and restore error message afterwards
-with something like I<prog_err_stderr(3)>.
+Parses one line from a properties file. C<path> is the path of the
+properties file. C<line> is the text to parse. C<lineno> is the current line
+number. The property parsed, if any, is added to C<map>. Emits error
+messages when syntax errors occur. That's probably a mistake. To suppress
+the error messages, call I<prog_err_none(3)> first and restore error
+messages afterwards with something like I<prog_err_stderr(3)>.
 
 */
 
@@ -297,7 +328,7 @@ static void prop_parse(Map *map, const char *path, char *line, size_t lineno)
 	/* Find first unquoted '=' */
 
 	for (p = cstr(prop), eq = strchr(p, '='); eq; eq = strchr(eq + 1, '='))
-		if (eq != p && eq[-1] != '\\')
+		if (eq == p || eq[-1] != '\\')
 			break;
 
 	if (!eq)
@@ -311,10 +342,23 @@ static void prop_parse(Map *map, const char *path, char *line, size_t lineno)
 	/* Identify and separate the name and value */
 
 	value = eq + 1;
-	while (eq > p && isspace((int)(unsigned char)eq[-1]))
+
+	while (is_space(value[0]))
+		++value;
+
+	while (eq > p && is_space(eq[-1]) && (eq == p + 1 || eq[-2] != '\\'))
 		--eq;
+
 	*eq = nul;
 
+	/* Unquote any quoted trailing space in the key */
+
+	if (eq > p + 1 && is_space(eq[-1]) && eq[-2] == '\\')
+	{
+		eq[-2] = eq[-1];
+		eq[-1] = nul;
+	}
+
 	/* Unquote any quoted '=' in the name */
 
 	if (!(name = unquote_equals(p)))
@@ -324,19 +368,15 @@ static void prop_parse(Map *map, const char *path, char *line, size_t lineno)
 		return;
 	}
 
-	/* Add this property to the map */
-
-	if (str_length(name) <= 0)
-	{
-		error("prop: %s line %d: Empty name\n%s", path, lineno, line);
-		str_release(prop);
-		str_release(name);
-		set_errno(EINVAL);
-		return;
-	}
-
 	key = cstr(name);
 
+	/* Unquote any quoted leading space in the value */
+
+	if (*value == '\\')
+		++value;
+
+	/* Add this property to the map */
+
 	if (!(val = mem_strdup(value)))
 	{
 		error("prop: Out of memory");
@@ -346,7 +386,10 @@ static void prop_parse(Map *map, const char *path, char *line, size_t lineno)
 	}
 
 	if (map_add(map, key, val) == -1)
+	{
 		error("prop: %s line %d: Property %s already defined\n%s", path, lineno, name, line);
+		mem_release(val);
+	}
 
 	str_release(prop);
 	str_release(name);
@@ -357,8 +400,8 @@ static void prop_parse(Map *map, const char *path, char *line, size_t lineno)
 C<Prop *prop_load(const char *path, Prop *defaults)>
 
 Creates and returns a new I<Prop> containing C<defaults> and the properties
-obtained from the properties file named C<path>. On error, returns C<null>
-with C<errno> set appropriately.
+obtained from the properties file specified by C<path>. On error, returns
+C<null> with C<errno> set appropriately.
 
 */
 
@@ -389,18 +432,24 @@ static Prop *prop_load(const char *path, Prop *defaults)
 
 C<int prop_init(void)>
 
-Initialises the L<prop(3)|prop(3)> module. Loads properties from the following locations:
+Initialises the I<prop(3)> module. Loads properties from the following locations:
 
-    /etc/properties/app             - system-wide, generic properties
-    $HOME/.properties/app           - user-defined, generic properties
-    /etc/properties/app.progname    - system-wide, program-specific properties
-    $HOME/.properties/app.progname  - user-defined, program-specific properties
+    /etc/properties/app          - system-wide, generic properties
+    ~/.properties/app            - user-defined, generic properties
+    /etc/properties/app.progname - system-wide, program-specific properties
+    ~/.properties/app.progname   - user-defined, program-specific properties
 
-Properties from the first three files become (nesting) defaults. Properties
-from the last file becomes the top level I<Prop>. If the last file doesn't
-exist, an empty I<Prop> becomes the top level I<Prop>.
+Note that, depending on how I<libslack> was installed, the system-wide
+C</etc/properties> directory might not be under C</etc>. It might be under
+another directory, such as C</usr/local/etc> or C</opt/local/etc>. If you
+aren't sure where it is expected, run C<libslack-config --cflags> and look
+for the definition of C<ETC_DIR>.
 
-Called at the start of the first get, set or unset function called.
+Properties from the first three files become (nested) defaults. Properties
+from the last file becomes the top-level I<Prop>. If the last file doesn't
+exist, an empty I<Prop> becomes the top-level I<Prop>.
+
+Called at the start of the first get, set, or unset function called.
 
 */
 
@@ -424,7 +473,7 @@ static int prop_init(void)
 	if ((prop_next = prop_load(path, prop)))
 		prop = prop_next;
 
-	/* User defined generic properties: $HOME/.properties/app */
+	/* User defined generic properties: ~/.properties/app */
 
 	if ((home = user_home()))
 	{
@@ -453,7 +502,7 @@ static int prop_init(void)
 		if ((prop_next = prop_load(path, prop)))
 			prop = prop_next;
 
-		/* User defined program specific properties: $HOME/.properties/app.progname */
+		/* User defined program specific properties: ~/.properties/app.progname */
 
 		if (home)
 		{
@@ -540,7 +589,7 @@ const char *prop_get(const char *name)
 =item C<const char *prop_get_or(const char *name, const char *default_value)>
 
 Returns the value of the property named C<name>. Returns C<default_value> on
-error or if there is no such property.
+error, or if there is no such property.
 
 =cut
 
@@ -557,11 +606,11 @@ const char *prop_get_or(const char *name, const char *default_value)
 
 =item C<const char *prop_set(const char *name, const char *value)>
 
-Sets the property named C<name> to a copy of C<value>. If I<prop_save(3)> is
-called after a call to this function, the new property will be saved to disk
-and will be available the next time this program is executed. On success,
-returns the copy of C<value>. On error, returns C<null> with C<errno> set
-appropriately.
+Sets the property named C<name> to a dynamically allocated copy of C<value>.
+If I<prop_save(3)> is called after a call to this function, the new property
+will be saved to disk, and will be available the next time this program is
+executed. On success, returns the copy of C<value>. On error, returns
+C<null> with C<errno> set appropriately.
 
 =cut
 
@@ -607,8 +656,8 @@ const char *prop_set(const char *name, const char *value)
 =item C<int prop_get_int(const char *name)>
 
 Returns the value of the property named C<name> as an integer. Returns C<0>
-on error or if there is no such property or if it is not interpretable as a
-decimal integer.
+on error, or if there is no such property, or if it is not interpretable as
+a decimal integer.
 
 =cut
 
@@ -624,7 +673,7 @@ int prop_get_int(const char *name)
 =item C<int prop_get_int_or(const char *name, int default_value)>
 
 Returns the value of the property named C<name> as an integer. Returns
-C<default_value> on error or if there is no such property or if it is not
+C<default_value> on error, or if there is no such property, or if it is not
 interpretable as a decimal integer.
 
 =cut
@@ -655,7 +704,7 @@ returns C<value>. On error, returns C<0>.
 int prop_set_int(const char *name, int value)
 {
 	char buf[128];
-	snprintf(buf, 128, "%d", value);
+	snprintf(buf, sizeof(buf), "%d", value);
 	return prop_set(name, buf) ? value : 0;
 }
 
@@ -664,8 +713,8 @@ int prop_set_int(const char *name, int value)
 =item C<double prop_get_double(const char *name)>
 
 Returns the value of the property named C<name> as a double. Returns C<0.0>
-on error or if there is no such property or if it is not interpretable as a
-floating point number.
+on error, or if there is no such property, or if it is not interpretable as
+a floating point number.
 
 =cut
 
@@ -681,7 +730,7 @@ double prop_get_double(const char *name)
 =item C<double prop_get_double_or(const char *name, double default_value)>
 
 Returns the value of the property named C<name> as a double. Returns
-C<default_value> on error or if there is no such property or if it is not
+C<default_value> on error, or if there is no such property, or if it is not
 interpretable as a floating point number.
 
 =cut
@@ -739,7 +788,7 @@ int prop_get_bool(const char *name)
 =item C<int prop_get_bool_or(const char *name, int default_value)>
 
 Returns the value of the property named C<name> as a boolean value. Returns
-C<default_value> on error or if there is no such property or if it is not
+C<default_value> on error, or if there is no such property, or if it is not
 interpretable as a boolean value. The values: C<"true">, C<"yes">, C<"on">
 and C<"1"> are all interpreted as C<true>. The values: C<"false">, C<"no">,
 C<"off"> and C<"0"> are all interpreted as C<false>. All other values are
@@ -761,7 +810,7 @@ int prop_get_bool_or(const char *name, int default_value)
 	if (sscanf(prop, " %d ", &val))
 		return val;
 
-	if (sscanf(prop, " %s ", buf))
+	if (sscanf(prop, " %127s ", buf))
 	{
 		if ((buf[0] == 't' || buf[0] == 'T') &&
 			(buf[1] == 'r' || buf[1] == 'R') &&
@@ -865,19 +914,19 @@ int prop_unset(const char *name)
 
 =item C<int prop_save(void)>
 
-Saves the program's properties to disk. If the C<"save"> property is set to
-C<"false">, C<"no">, C<"off"> or C<"0">, nothing is written to disk. If no
-properties were added, removed or changed, nothing is written to disk. Only
-the user-defined, program-specific properties are saved. Generic and
+Saves the program's properties to disk. If the property named C<"save"> is
+set to C<"false">, C<"no">, C<"off"> or C<"0">, nothing is written to disk.
+If no properties were added, removed or changed, nothing is written to disk.
+Only the user-defined, program-specific properties are saved. Generic and
 system-wide properties files must be edited by hand. Each program can only
 save its own properties. They are saved in the following file:
 
-    $HOME/.properties/app.progname
+    ~/.properties/app.progname
 
 Where C<progname> is the name of the program after being translated as
-described at the top of the DESCRIPTION section.
+described at the top of the I<DESCRIPTION> section.
 
-The F<.properties> directory is created if necessary. On success, returns
+The C<~/.properties> directory is created if necessary. On success, returns
 C<0>. On error, returns C<-1> with C<errno> set appropriately.
 
 =cut
@@ -1022,6 +1071,21 @@ int prop_save(void)
 		str_release(lhs);
 		lhs = lhs2;
 
+		/* Quote any trailing space in the key */
+
+		if (str_length(lhs) && is_space(cstr(lhs)[str_length(lhs) - 1]))
+		{
+			if (!str_insert(lhs, str_length(lhs) - 1, "\\"))
+			{
+				fclose(file);
+				lister_release(k);
+				list_release(keys);
+				str_release(lhs);
+				locker_unlock(g.locker);
+				return -1;
+			}
+		}
+
 		/* Quote any special chars in the value */
 
 		if (!(rhs = quote_special(value)))
@@ -1034,6 +1098,21 @@ int prop_save(void)
 			return -1;
 		}
 
+		/* Quote any leading space in the value */
+
+		if (is_space(cstr(rhs)[0]))
+		{
+			if (!str_insert(rhs, 0, "\\"))
+			{
+				fclose(file);
+				lister_release(k);
+				list_release(keys);
+				str_release(lhs);
+				locker_unlock(g.locker);
+				return -1;
+			}
+		}
+
 		fprintf(file, "%s=%s\n", cstr(lhs), cstr(rhs));
 		str_release(lhs);
 		str_release(rhs);
@@ -1083,9 +1162,9 @@ int prop_clear(void)
 
 =item C<int prop_locker(Locker *locker)>
 
-Sets the locking strategy for the L<prop(3)|prop(3)> module to C<locker>.
-This is only needed in multi threaded programs. It must only be called once,
-from the main thread. On success, returns C<0>. On error, returns C<-1> with
+Sets the locking strategy for the I<prop(3)> module to C<locker>. This is
+only needed in multi-threaded programs. It must only be called once, from
+the main thread. On success, returns C<0>. On error, returns C<-1> with
 C<errno> set appropriately.
 
 =cut
@@ -1112,25 +1191,25 @@ On error, C<errno> is set either by an underlying function, or as follows:
 
 =over 4
 
-=item EINVAL
+=item C<EINVAL>
 
-When there is no prog_name or home directory or when there is a parse error
-in a properties file or if the F<~/.properties> exists but is not a
-directory. I<prop_locker(3)> sets this when an attempt is made to change the
-locker that has already been set.
+When there is no prog_name or home directory, or when there is a parse error
+in a properties file, or if C<~/.properties> exists but is not a directory.
+I<prop_locker(3)> sets this when an attempt is made to change the locker
+after one has already been set.
 
 =back
 
 =head1 MT-Level
 
-MT-Disciplined
+I<MT-Disciplined>
 
 =head1 FILES
 
     /etc/properties/app
-    $HOME/.properties/app
+    ~/.properties/app
     /etc/properties/app.*
-    $HOME/.properties/app.*
+    ~/.properties/app.*
 
 =head1 EXAMPLE
 
@@ -1164,17 +1243,19 @@ MT-Disciplined
 
 =head1 BUGS
 
-This only provides coarse grained persistence. If multiple instances of the
-same program are setting properties, the last to exit wins.
+This only provides coarse-grained persistence. If multiple instances of the
+same program are setting properties, the last to exit wins. This can be
+overcome by calling I<prop_save()> after setting any property and
+I<prop_clear()> before getting any property.
 
 =head1 SEE ALSO
 
-L<libslack(3)|libslack(3)>,
-L<prog(3)|prog(3)>
+I<libslack(3)>,
+I<prog(3)>
 
 =head1 AUTHOR
 
-20020916 raf <raf@raf.org>
+20230819 raf <raf@raf.org>
 
 =cut
 
@@ -1271,10 +1352,11 @@ int main(int ac, char **av)
 	data[] =
 	{
 		{ "key", "value" },
-		{ "key with spaces", "value with spaces" },
+		{ " key with spaces ", "value with spaces" },
 		{ "key with = sign", " value with leading space" },
 		{ "key with newline\n and = two = signs", "value with newline\n!" },
 		{ "key with newline,\n = two = signs and an Escape\033!", "value with newline\n and two non printables\001!\002!" },
+		{ "", "" },
 		{ NULL, NULL }
 	};
 
@@ -1285,6 +1367,7 @@ int main(int ac, char **av)
 	const char *val;
 	int ival, i;
 	int int_val;
+	const char *sval;
 	double double_val;
 	int bool_val;
 	int errors = 0;
@@ -1302,8 +1385,8 @@ int main(int ac, char **av)
 	if (strcmp(val, value))
 		++errors, printf("Test1: prop_set(key, value) failed (%s not %s)\n", val, value);
 	val = prop_get(key);
-	if (strcmp(val, value))
-		++errors, printf("Test2: prop_get(key) failed (%s not %s)\n", val, value);
+	if (!val || strcmp(val, value))
+		++errors, printf("Test2: prop_get(key) failed (%s not %s)\n", val ? val : "null", value);
 	val = prop_get_or(key, NULL);
 	if (strcmp(val, value))
 		++errors, printf("Test3: prop_get_or(key, NULL) failed (%s not %s)\n", val, value);
@@ -1337,7 +1420,7 @@ int main(int ac, char **av)
 		++errors, printf("Test11: prop_set(key, value) failed (%s not %s)\n", val, value);
 	ival = prop_save();
 	if (ival != 0)
-		++errors, printf("Test12: prop_save() (with progname) failed (%d not 0)\n", ival);
+		++errors, printf("Test12: prop_save() (with progname) failed (%d not 0, errno = %s)\n", ival, strerror(errno));
 
 	clean(has_props);
 	prop_clear();
@@ -1346,66 +1429,124 @@ int main(int ac, char **av)
 	{
 		val = prop_set(data[i].key, data[i].value);
 		if (strcmp(val, data[i].value))
-			++errors, printf("Test%d: prop_set('%s', '%s') failed ('%s' not '%s')\n", 13 + 2 * i, data[i].key, data[i].value, val, value);
+			++errors, printf("Test%d: prop_set('%s', '%s') failed ('%s' not '%s')\n", 13 + 2 * i, data[i].key, data[i].value, val, data[i].value);
 		val = prop_get(data[i].key);
-		if (strcmp(val, data[i].value))
-			++errors, printf("Test%d: prop_get('%s') failed ('%s' not '%s')\n", 14 + 2 * i, data[i].key, val, value);
+		if (!val || strcmp(val, data[i].value))
+			++errors, printf("Test%d: prop_get('%s') failed ('%s' not '%s')\n", 14 + 2 * i, data[i].key, val ? val : "null", data[i].value);
 	}
 
 	ival = prop_save();
 	if (ival != 0)
-		++errors, printf("Test21: prop_save() (with progname) failed (%d not 0)\n", ival);
+		++errors, printf("Test25: prop_save() (with progname) failed (%d not 0, errno = %s)\n", ival, strerror(errno));
 
 	prop_clear();
 
 	for (i = 0; data[i].key; ++i)
 	{
 		val = prop_get(data[i].key);
-		if (strcmp(val, data[i].value))
-			++errors, printf("Test%d: prop_get('%s') failed ('%s' not '%s')\n", 22 + i, data[i].key, val, data[i].value);
+		if (!val || strcmp(val, data[i].value))
+			++errors, printf("Test%d: prop_get('%s') failed ('%s' not '%s')\n", 26 + i, data[i].key, val ? val : "null", data[i].value);
 	}
 
 	clean(has_props);
 	prop_clear();
 
 	if ((int_val = prop_set_int("i", 37)) != 37)
-		++errors, printf("Test26: prop_set_int() failed (%d not 37)\n", int_val);
+		++errors, printf("Test32: prop_set_int() failed (%d not 37)\n", int_val);
 
 	if ((int_val = prop_get_int("i")) != 37)
-		++errors, printf("Test27: prop_get_int() failed (%d not 37)\n", int_val);
+		++errors, printf("Test33: prop_get_int() failed (%d not 37)\n", int_val);
 
 	if ((int_val = prop_get_int_or("i", 13)) != 37)
-		++errors, printf("Test28: prop_get_int_or() failed (%d not 37)\n", int_val);
+		++errors, printf("Test34: prop_get_int_or() failed (%d not 37)\n", int_val);
 
 	if ((int_val = prop_get_int_or("j", 13)) != 13)
-		++errors, printf("Test29: prop_get_int_or() failed (%d not 13)\n", int_val);
+		++errors, printf("Test35: prop_get_int_or() failed (%d not 13)\n", int_val);
 
 	if ((double_val = prop_set_double("d", 37.0)) != 37.0)
-		++errors, printf("Test30: prop_set_double() failed (%g not 37.0)\n", double_val);
+		++errors, printf("Test36: prop_set_double() failed (%g not 37.0)\n", double_val);
 
 	if ((double_val = prop_get_double("d")) != 37)
-		++errors, printf("Test31: prop_get_double() failed (%g not 37.0)\n", double_val);
+		++errors, printf("Test37: prop_get_double() failed (%g not 37.0)\n", double_val);
 
 	if ((double_val = prop_get_double_or("d", 13.0)) != 37)
-		++errors, printf("Test32: prop_get_double_or() failed (%g not 37.0)\n", double_val);
+		++errors, printf("Test38: prop_get_double_or() failed (%g not 37.0)\n", double_val);
 
 	if ((double_val = prop_get_double_or("e", 13.0)) != 13.0)
-		++errors, printf("Test33: prop_get_double_or() failed (%g not 13.0)\n", double_val);
+		++errors, printf("Test39: prop_get_double_or() failed (%g not 13.0)\n", double_val);
 
 	if ((bool_val = prop_set_bool("b", 1)) != 1)
-		++errors, printf("Test34: prop_set_bool() failed (%d not 1)\n", bool_val);
+		++errors, printf("Test40: prop_set_bool() failed (%d not 1)\n", bool_val);
 
 	if ((bool_val = prop_get_bool("b")) != 1)
-		++errors, printf("Test35: prop_get_bool() failed (%d not 1)\n", bool_val);
+		++errors, printf("Test41: prop_get_bool() failed (%d not 1)\n", bool_val);
 
 	if ((bool_val = prop_get_bool_or("b", 0)) != 1)
-		++errors, printf("Test36: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+		++errors, printf("Test42: prop_get_bool_or() failed (%d not 1)\n", bool_val);
 
 	if ((bool_val = prop_get_bool_or("c", 1)) != 1)
-		++errors, printf("Test37: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+		++errors, printf("Test43: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+
+	sval = prop_set("b", "true");
+	if (!sval)
+		++errors, printf("Test44: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "true"))
+		++errors, printf("Test45: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "true");
+	if ((bool_val = prop_get_bool_or("b", 0)) != 1)
+		++errors, printf("Test46: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+
+	sval = prop_set("b", "false");
+	if (!sval)
+		++errors, printf("Test47: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "false"))
+		++errors, printf("Test48: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "true");
+	if ((bool_val = prop_get_bool_or("b", 1)) != 0)
+		++errors, printf("Test49: prop_get_bool_or() failed (%d not 0)\n", bool_val);
+
+	sval = prop_set("b", "yes");
+	if (!sval)
+		++errors, printf("Test50: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "yes"))
+		++errors, printf("Test51: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "yes");
+	if ((bool_val = prop_get_bool_or("b", 0)) != 1)
+		++errors, printf("Test52: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+
+	sval = prop_set("b", "no");
+	if (!sval)
+		++errors, printf("Test53: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "no"))
+		++errors, printf("Test54: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "no");
+	if ((bool_val = prop_get_bool_or("b", 1)) != 0)
+		++errors, printf("Test55: prop_get_bool_or() failed (%d not 0)\n", bool_val);
+
+	sval = prop_set("b", "on");
+	if (!sval)
+		++errors, printf("Test56: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "on"))
+		++errors, printf("Test57: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "on");
+	if ((bool_val = prop_get_bool_or("b", 0)) != 1)
+		++errors, printf("Test58: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+
+	sval = prop_set("b", "off");
+	if (!sval)
+		++errors, printf("Test59: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "off"))
+		++errors, printf("Test60: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "off");
+	if ((bool_val = prop_get_bool_or("b", 1)) != 0)
+		++errors, printf("Test61: prop_get_bool_or() failed (%d not 0)\n", bool_val);
+
+	sval = prop_set("b", "neither");
+	if (!sval)
+		++errors, printf("Test62: prop_set() failed (returned null)\n");
+	else if (strcmp(sval, "neither"))
+		++errors, printf("Test63: prop_set() failed (returned \"%s\" not \"%s\")\n", sval, "neither");
+	if ((bool_val = prop_get_bool_or("b", 1)) != 1)
+		++errors, printf("Test64: prop_get_bool_or() failed (%d not 1)\n", bool_val);
+	if ((bool_val = prop_get_bool_or("b", 0)) != 0)
+		++errors, printf("Test65: prop_get_bool_or() failed (%d not 0)\n", bool_val);
 
 	if (errors)
-		printf("%d/37 tests failed\n", errors);
+		printf("%d/65 tests failed\n", errors);
 	else
 		printf("All tests passed\n");
 
